/**
 * @param {Array<Event>} events
 * @returns {Array<Event>}
 */
export function postprocess(
  events: Array<import('micromark-util-types').Event>
): Array<import('micromark-util-types').Event>
export type Event = import('micromark-util-types').Event
